/*
# WebSocket IRC with HTTP-based website interface

This script is the v5 shoutbox IRC client. Unlike the legacy v4 shoutbox which
was a poll-based HTTP service, this shoutbox is run by an IRC server, hence the
shoutbox is powered by this IRC client.

Support for IRC features is fairly limited as the shoutbox has never been a
very fancy chat. For full functionality, use a standard IRC client.


## Posting via HTTP to use the Planète Casio session cookie

The one unusual feature in this program is the interaction with the Planète
Casio website. This solves an issue of permissions that is difficult to manage
with pure IRC. The expectation for the shoutbox is that:
- Anyone can read but only community members can write;
- It should load immediately when embedded into website pages and allow logged
  in users to read and write messages with no further interactions.

In the v4's cursed design, the shoutbox was part of the website API, so users
logged into their Planète Casio account would be recognized by the shoutbox
server (which is just the website) and authenticated by their session cookie.
This doesn't work with the IRC server, since the connections are very different
and cookies are not a reliable method of authentication. With a standard IRC
client, users would need to type in their password to connect to the chat
despite already being logged in to the website.

The chosen solution is as follows:
- Every embedded shoutbox connects to the IRC server as a read-only guest;
- Sending a message is done via a website request. The website identifies
  logged in users by their session cookies and forwards the messages to the
  IRC server locally;
- Messages from the website to the IRC server are attributed to their original
  authors by whichever nick/plugin trick works best.

(WIP at time of writing)


## Message protocol

This IRC client usually runs inside of a service worker to that it can be
shared between any number of open Planète Casio tabs. Thus, communication
between this client and the user interface uses a message-based interface,
which is intended to work like RPC.
*/

"use strict";

class IRCMessage {
  /* Regex for parsing messages */
  static #rMessage = (function(){
    const rTags     = /(?:@([^ ]+) +)?/;
    const rSource   = /(?::([^ ]+) +)?/;
    const rCommand  = /(?:([a-zA-Z]+)|(\d{3}))/;
    const rArgs     = /((?: +[^\r\n :][^\r\n ]*)*)/;
    const rTrailing = /(?: +:([^\r\n]*))?/;
    return new RegExp([/^/, rTags, rSource, rCommand, rArgs, rTrailing, /$/]
                      .map(r => r.source).join(""));
  })();

  constructor(text) {
    const matches = IRCMessage.#rMessage.exec(text);
    if(matches === undefined)
      return;

    this.tags = new Map();
    (matches[1] || "").split(";").filter(s => s).forEach(s => {
      const parts = s.split(/=(.*)/);
      const value = (parts[1] || "").replaceAll("\\\\", "\\")
                                    .replaceAll("\\:", ";")
                                    .replaceAll("\\r", "\r")
                                    .replaceAll("\\n", "\n")
                                    .replaceAll(/\\./g, m => m[1]);
      this.tags.set(parts[0], value);
    });

    this.source = matches[2];
    this.command = matches[3] || parseInt(matches[4]);
    this.args = (matches[5] || "").split(" ").filter(s => s);
    if(matches[6] !== undefined)
      this.args.push(matches[6]);
  }
}

const State = Object.freeze({
  /* Websocket is disconnected. */
  DISCONNECTED: Symbol("DISCONNECTED"),
  /* We just opened a socket to the IRC server and have yet to register. */
  CONNECTING: Symbol("CONNECTING"),
  /* We are authenticated and ready to use the server. */
  READY: Symbol("READY"),
});

/* https://stackoverflow.com/questions/30106476 */
function base64_encode(str) {
  return btoa(encodeURIComponent(str).replace(/%([0-9A-F]{2})/g,
      function toSolidBytes(match, p1) {
          return String.fromCharCode('0x' + p1);
  }));
}

/* Clonable IRC state passed to user with most messages. Users keep a copy of
   this object around so they can display the state without communicating with
   the IRC client (which is generally in another thread) all the time. */
export class ClientState {
  constructor(client) {
    this.state = client.state;
    this.channels = client.channels;
  }

  /* Connection status */
  isConnected() {
    return this.state !== State.DISCONNECTED;
  }
  isRunning() {
    return this.state === State.READY;
  }
  stateString() {
    switch(this.state) {
    case State.DISCONNECTED:    return "Disconnected";
    case State.CONNECTING:      return "Connecting...";
    case State.READY:           return "Connected";
    }
  }
}

export class Client {
  #ws;
  #password;

  constructor(serverUrl, postUrl) {
    /* IRC server's websocket URL (wss://...). */
    this.serverUrl = serverUrl;
    /* HTTP-based message post URL. Undefined when using pure IRC, or a v4/v5
       URL when using website authentication. */
    this.postUrl = postUrl;
    /* Websocket for IRC server communication. */
    this.#ws = undefined;
    /* Current connection state. */
    this.state = State.DISCONNECTED;
    /* Capabilities enabled. */
    this.caps = [];
    /* List of channels and their information */
    this.channels = new Map();
    /* Username with which we are logged in. Undefined when DISCONNECTED. */
    this.username = undefined;
    /* Password for connection. Erased after authenticating. */
    this.#password = undefined;
    /* Batches currently being processed */
    this.batches = new Map();
  }

  /*** Callback messaging ***/

  onMessage() {}
  sendMessage(e) {
    e.state = new ClientState(this);
    this.onMessage(e);
  }

  log(text) {
    this.sendMessage({ type: "log", text: text });
  }
  onStateChanged() {
    this.sendMessage({ type: "state_changed" });
  }
  onAuthenticated() {
    this.sendMessage({ type: "authenticated" });
  }
  onNewMessage(channel, date, author, message) {
    this.sendMessage({
      type: "new_message",
      channel: channel,
      date: date,
      author: author,
      message: message,
    });
  }
  onChannelChanged(channel) {
    this.sendMessage({ type: "channel_changed", channel: channel });
  }

  /*** Message-based actions ***/

  remoteCall(e) {
    if(e.type == "connect")
      this.connect(e.login, e.password);
    else if(e.type == "post")
      this.postMessage(e.channel, e.message);
  }

  /*** Connection management ***/

  setState(state) {
    this.state = state;
    this.onStateChanged();
  }

  connect(username, password) {
    let ws = new WebSocket(this.serverUrl);
    this.#ws = ws;
    this.log("[v5shoutbox] Connecting...");

    ws.onopen = function() {
      this.log("[v5shoutbox] Connected.");
      this.setState(State.CONNECTING);
      this.username = username;
      this.#password = password;
      this.#startExchange();
    }.bind(this);

    ws.onmessage = function(net) {
      this.log("[<] " + net.data);
      let msg = new IRCMessage(net.data);
      if(msg.command === undefined) {
        this.log("[v5shoutbox] invalid message");
        return;
      }
      this.processIRCMessage(msg);
    }.bind(this);

    ws.onclose = function() {
      this.log("[v5shoutbox] Disconnected.");
      this.setState(State.DISCONNECTED);
    }.bind(this);
  }

  #sendRaw(command) {
    this.log("[>] " + command);
    this.#ws.send(command);
  }

  send(command, args) {
    if(args.slice(0,-1).some(x => x.includes(" ")))
      return this.log("invalid spaces in non-last args: " + command.toString() +
                 " " + args.toString());

    if(args.length && args[args.length - 1].includes(" "))
      args[args.length - 1] = ":" + args[args.length - 1];

    this.#sendRaw(command + " " + args.join(" "));
  }

  /*** IRC protocol ***/

  #startExchange() {
    this.#sendRaw("CAP LS 302");
    this.send("NICK", [this.username]);
    this.send("USER", [this.username, "*", "0", this.username]);
  }

  processIRCMessage(msg) {
    if(msg.command === "PING") {
      this.send("PONG", msg.args);
      return;
    }
    if(msg.command === "BATCH") {
      const [mode, id] = [msg.args[0][0], msg.args[0].substring(1)];
      if(mode == "+")
        this.batches.set(id, []);
      this.batches.get(id).push(msg);
      if(mode == "-") {
        this.processBatch(this.batches.get(id));
        this.batches.delete(id);
      }
      return;
    }
    if(msg.tags.has("batch")) {
      this.batches.get(msg.tags.get("batch")).push(msg);
      return;
    }

    if(msg.command === "CAP" && msg.args[1] === "LS") {
      const caps = msg.args[2].split(" ").filter(x => x);
      const wishlist =
        ["message-tags", "server-time", "echo-message",
         "batch", "sasl", "draft/chathistory"];

      let req = wishlist.filter(x => caps.includes(x));
      if(caps.some(x => x.startsWith("sasl=")))
        req.push("sasl");

      if(req.length)
        this.send("CAP", ["REQ", req.join(" ")]);
      else {
        this.caps = [];
        this.#sendRaw("CAP END");
      }
    }

    if(msg.command === "CAP" && msg.args[1] === "ACK") {
      this.caps = msg.args[2].split(" ").filter(x => x);
      if(this.caps.includes("sasl"))
        this.#sendRaw("AUTHENTICATE PLAIN");
      else
        this.#sendRaw("CAP END");
    }
    if(msg.command === "AUTHENTICATE" && msg.args[0] === "+") {
      const arg = this.username + "\0" + this.username + "\0" + this.#password;
      const b64 = base64_encode(arg);
      this.log("[v5shoutbox] AUTHENTICATE command sent (not shown)");
      this.#ws.send("AUTHENTICATE " + b64);
    }

    /* Default authentication if SASL is unavailable or fails */
    if(msg.command === "NOTICE" && msg.args[1].includes("/AUTH")) {
      this.log("[v5shoutbox] AUTH command sent (not shown)");
      this.#ws.send("AUTH " + this.username + ":" + this.#password);
    }

    if(msg.command === 900) {
      this.log("[v5shoutbox] Authenticated.");
      if(this.caps.includes("sasl"))
        this.#sendRaw("CAP END");
      this.setState(State.READY);
      this.#password = undefined;
      this.onAuthenticated();

      if(this.caps.includes("draft/chathistory")) {
        ["#annonces", "#projets", "#hs"].forEach(channel => {
          this.send("CHATHISTORY", ["LATEST", channel, "*", "100"]);
        });
      }
    }

    if(msg.command === "JOIN"
       && msg.tags.get("account") === this.username) {
      this.channels.set(msg.args[0], {});
      this.onChannelChanged(msg.args[0]);
    }
    if(msg.command === 332) {
      this.channels.get(msg.args[1]).header = msg.args[2];
      this.onChannelChanged(msg.args[1]);
    }
    if(msg.command === 353 && msg.args[1] === "=") {
      this.channels.get(msg.args[2]).clients = msg.args[3].split(" ");
      this.onChannelChanged(msg.args[2]);
    }

    if(msg.command === "PRIVMSG")
      this.processPRIVMSG(msg);
  }

  processBatch(batch) {
    const batchArgs = batch[0].args.slice(1);
    if(batchArgs.length === 2 && batchArgs[0] === "chathistory") {
      batch.forEach(msg => {
        if(msg.command === "PRIVMSG")
          this.processPRIVMSG(msg);
      });
    }
  }

  processPRIVMSG(msg) {
    if(msg.args.length == 2) {
      let source = msg.source;
      if(source.includes("!"))
        source = source.substr(0, source.indexOf("!"));
      let date = new Date();
      if(msg.tags.has("time"))
        date = new Date(msg.tags.get("time"));
      this.onNewMessage(msg.args[0], date, source, msg.args[1]);
    }
  }

  postMessage(channel, message) {
    if(message.startsWith("/")) {
      this.#sendRaw(message.substring(1));
    }
    else {
      this.send("PRIVMSG", [channel, message]);
      if(!this.caps.includes("echo-message"))
        this.onNewMessage(channel, new Date(), conn.username, message);
    }
  }
}
